import ctypes
import math
import hashlib
import re
import os

import c4d
import geom
from functools import reduce


DEBUG = False
FRACTURE_ID = 1018791
TAG_ID = 1028660

#=====================================================================================================================#
#   SYMBOLS
#=====================================================================================================================#
try:
    import symbol_parser
    c4d_symbol_file = os.path.join(os.path.dirname(__file__), "../res")
    symbol_parser.parse_and_export_in_caller(c4d_symbol_file)
except ImportError:
    ART_FILE		=30000
    ART_LAYEROFFSET		=30001
    ART_DEPTH		=30002
    ART_CONTENTS		=30003
    ART_RECENTER		=30004
    ART_PATHOFFSET		=30005
    ART_USEMOGRAPH		=30006
    ART_CINEWARE		=30007
    ART_FILTERGROUP		=30010
    ART_LAYERFILTER		=30011
    ART_PATHFILTER_ON	=30012
    ART_PATHFILTER		=30013
    ART_RECENTER_Z		=30014
    ART_FORCEPDF		=30015
    ART_STROKESETTINGS	=30016
    ART_STROKESCALE		=30017
    ART_STROKEASPECT	=30018
    ART_STROKEROUNDING	=30019
    ART_STROKEGROWTH	=30020
    ART_STROKEALIGN		=30021
    ART_STROKEALIGN_START = 30022
    ART_STROKEALIGN_END = 30023
    ART_STROKEALIGN_CENTER = 30024
    ART_STROKEALIGN_STARTEND = 30025
    ART_STROKEOFFSET = 30026
    ART_GREEDYFILL = 30050
    ART_AUTOCLOSE = 30051
    ART_LEGACYCENTER = 30052
    ART_PATHEFFECTOR	=30100
    ART_PATHEFFECTOR_LIST = 30101
    ART_LAYEREFFECTOR   =30200
    ART_LAYEREFFECTOR_LIST = 30201
    ART_RELOAD		=5000
    ART_OPEN		=5001
    ART_CREATEMATERIALS		=5002
    ART_INFO		=5005
    ART_DEBUGMSG	=5006
    ART_CREATEOBJECT = 6000
    ART_CHECKSUM		=9000
    ART_STROKED			=9001
    ART_FILLED			=9002
    ART_STROKECOLOR		=9003
    ART_FILLCOLOR		=9004
    ART_STROKEWIDTH		=9005

checksums = {}
allpoints = []

def FindCenter(points):
    global allpoints

    maxVec = c4d.Vector(points[0].x,points[0].y,points[0].z)
    minVec = c4d.Vector(points[0].x,points[0].y,points[0].z)

    for i in range(len(points)):
        if points[i].x > maxVec.x: maxVec.x = points[i].x
        if points[i].y > maxVec.y: maxVec.y = points[i].y
        if points[i].z > maxVec.z: maxVec.z = points[i].z
        if points[i].x < minVec.x: minVec.x = points[i].x
        if points[i].y < minVec.y: minVec.y = points[i].y
        if points[i].z < minVec.z: minVec.z = points[i].z

    allpoints += points

    return ((maxVec - minVec) / float(2)) + minVec

def CreatePath(path,node=None,userdataIdx=[],layerName="",axis="Z"):
    if not path or not isinstance(path, geom.Path): return None,0,0,0
    
    points = []
    tangents = {}
    segments = []
    
    if DEBUG:
        print(path)
        print(path.name)
        print("Filled=",path.filled)
        print("FilledEvenOdd=",path.fillevenodd)
        print("Stroked=",path.stroked)
        print("FillPaint=",path.fillpaint.color)
        print("StrokePaint=",path.strokepaint.color)
        print("StrokeWidth=",path.width)
        print("Markers=",path.markers)

    # For some reason, C4D tangents are 3/4 the length of the bezier tangent
    # Found this on a plugincafe post - not sure why it's the case
    # Maybe a hermite / bezier thing?
    f = 0.75
        
    for sub in path.subpaths:
        if DEBUG:
            print(sub.segments)
        
        for seg in sub.segments:
            if seg[0] == "L":
                if not points or points[-1] != c4d.Vector(seg[1][0], seg[1][1], 0):
                    points.append(c4d.Vector(seg[1][0], seg[1][1], 0))
                    tangents[len(points)-1] = {"l":c4d.Vector(0,0,0), "r":c4d.Vector(0,0,0)}
                points.append(c4d.Vector(seg[2][0], seg[2][1], 0))
                tangents[len(points)-1] = {"l":c4d.Vector(0,0,0), "r":c4d.Vector(0,0,0)}
            elif seg[0] == "B":
                if not points or points[-1] != c4d.Vector(seg[1][0], seg[1][1], 0):
                    points.append(c4d.Vector(seg[1][0], seg[1][1], 0))
                    if len(points)-1 not in tangents: tangents[len(points)-1] = {}
                tangents[len(points)-1]["r"] = c4d.Vector(seg[3][0], seg[3][1], 0)-c4d.Vector(seg[1][0], seg[1][1], 0)
                points.append(c4d.Vector(seg[2][0], seg[2][1], 0))
                tangents[len(points)-1] = {"l":c4d.Vector(seg[4][0], seg[4][1], 0)-c4d.Vector(seg[2][0], seg[2][1], 0)}
            else:
                #Q & A not handled yet
                print("*"*5,seg[0]," NOT HANDLED")
                
        if points:

            closed = False
            start = 0
            if segments: start = segments[-1]["end"]
            if points[start] == points[len(points)-1]:
                points.pop()
                tangents[start]["l"] = tangents.pop(len(points))["l"]
                closed = True                        
            segments.append({"cnt":len(points)-start,"closed":closed,"start":start,"end":len(points)})

            if DEBUG:
                print("#"*10)
                for idx, p in enumerate(points):
                    print(idx,"::",p,"::",tangents[idx])
                print("#"*10)

    if points:

        chksum = GenerateChecksum(points)

        if axis == "Y":
            for i in range(len(points)):
                points[i].z = -points[i].y
                points[i].y = 0

            for i in range(len(tangents)):
                if "l" in tangents[i]:
                    tangents[i]["l"].z = -tangents[i]["l"].y
                    tangents[i]["l"].y = 0
                if "r" in tangents[i]:
                    tangents[i]["r"].z = -tangents[i]["r"].y
                    tangents[i]["r"].y = 0

        elif axis == "X":
            for i in range(len(points)):
                points[i].z = -points[i].x
                points[i].x = 0

            for i in range(len(tangents)):
                if "l" in tangents[i]:
                    tangents[i]["l"].z = -tangents[i]["l"].x
                    tangents[i]["l"].x = 0
                if "r" in tangents[i]:
                    tangents[i]["r"].z = -tangents[i]["r"].x
                    tangents[i]["r"].x = 0

        # Re-center axis
        if node and node[ART_LEGACYCENTER]:
            center = sum(points) / len(points)
        else:
            center = FindCenter(points)
        for i in range(len(points)):
            points[i] = points[i]-center
         
        oldchksum = _GenerateChecksum(points)
                
        exists = False
        
        name = "Path"
        if path.name: name = path.name
        
        if chksum in checksums:
            
            spl = checksums[chksum]
            pathinfo = spl.GetTag(TAG_ID)
            if pathinfo:
                if path.stroked:
                    pathinfo[ART_STROKED] = path.stroked
                    pathinfo[ART_STROKECOLOR] = c4d.Vector(path.strokepaint.color[0],path.strokepaint.color[1],path.strokepaint.color[2])
                    if path.width == 0:
                        path.width = 1
                        pathinfo[ART_STROKEWIDTH] = path.width
                if path.filled:
                    pathinfo[ART_FILLED] = path.filled
                    pathinfo[ART_FILLCOLOR] = c4d.Vector(path.fillpaint.color[0],path.fillpaint.color[1],path.fillpaint.color[2])
                        
            exists = True
        else:
        
            spl = c4d.SplineObject(len(points),c4d.SPLINETYPE_BEZIER)
            if spl is None: return None,0,0,0
            checksums[chksum] = spl
            
            spl.ResizeObject(len(points),len(segments))
            spl[c4d.SPLINEOBJECT_CLOSED] = closed
            if node and node[ART_AUTOCLOSE]:
                spl[c4d.SPLINEOBJECT_CLOSED] = True
            spl.SetAbsPos(center)

            spl.SetAllPoints(points)
            for pt, tan in tangents.items():
                if "l" not in tan: tan["l"] = c4d.Vector(0,0,0)
                if "r" not in tan: tan["r"] = c4d.Vector(0,0,0)
                spl.SetTangent(pt, tan["l"]*f, tan["r"]*f)
            
            for s, seg in enumerate(segments):
                spl.SetSegment(s,seg["cnt"],seg["closed"])
                
            spl.Message(c4d.MSG_UPDATE)
            
            col = 0,0,0
            if path.stroked: 
                #name += ".Stroked"
                col = path.strokepaint.color
            if path.filled: 
                #name += ".Filled"
                col = path.fillpaint.color       
            
            # Set Basic Parameters
            spl.SetName(name)
            spl[c4d.ID_BASEOBJECT_USECOLOR] = c4d.ID_BASEOBJECT_USECOLOR_ALWAYS
            spl[c4d.ID_BASEOBJECT_COLOR] = c4d.Vector(col[0],col[1],col[2])

            # Set Spline Interpolation from ArtSmartObject
            if node:
                for i,val in node.GetData():
                    if i>=1000 and i<2000:
                        spl[i]=val

            # Adjust for default stroke width of 1
            if path.stroked and path.width == 0:
                path.width = 1
                        
            # Create PathInfo tag
            pathinfo = spl.MakeTag(TAG_ID)
            pathinfo[ART_CHECKSUM] = chksum
            pathinfo[ART_STROKED] = path.stroked
            pathinfo[ART_FILLED] = path.filled
            pathinfo[ART_STROKECOLOR] = c4d.Vector(path.strokepaint.color[0],path.strokepaint.color[1],path.strokepaint.color[2])
            pathinfo[ART_FILLCOLOR] = c4d.Vector(path.fillpaint.color[0],path.fillpaint.color[1],path.fillpaint.color[2])
            pathinfo[ART_STROKEWIDTH] = path.width
            
        ext = None
        swp = None
        strokewidth = 0
        if node: strokewidth = node[ART_STROKESCALE] * path.width
        depth = GetUserDataValue(node,userdataIdx,"Dep",layerName,chksum)
        spread = GetUserDataValue(node,userdataIdx,"Pat",layerName,chksum)
        offset = GetUserDataValue(node,userdataIdx,"Off",layerName,chksum)
        if path.filled and depth != 0: 
            ext = CreateExtrude(spl,node,depth,pathinfo[ART_FILLCOLOR],userdataIdx,chksum,oldchksum)
            spl.InsertUnder(ext)
            spl.SetAbsPos(c4d.Vector(0))
            exists = True
        if path.stroked and strokewidth > 0:
            if exists:
                spl = spl.GetClone()
                spl.SetAbsPos(center)
            swp = CreateSweep(spl,node,pathinfo[ART_STROKECOLOR],userdataIdx,chksum,depth,strokewidth)
        
        # All Returns Here
        # Both stroked & filled - return a null containing extrude & sweep
        null = c4d.BaseObject(c4d.Onull)
        null.SetAbsPos(center)
        null.SetName(name)
        null.InsertTag(pathinfo.GetClone())
        if ext and swp:          
            swp.SetAbsPos(c4d.Vector(0))
            ext.SetAbsPos(c4d.Vector(0))
            swp.InsertUnder(null)
            ext.InsertUnder(null)
            return null,depth,offset,spread
        elif ext:
            # Just filled - return the extrusion
            ext.SetAbsPos(c4d.Vector(0))
            ext.InsertUnder(null)
            return null,depth,offset,spread
        elif swp:
            # Just stroked - return the sweep
            swp.SetAbsPos(c4d.Vector(0))
            swp.InsertUnder(null)
            return null,depth,offset,spread
        elif exists: 
            # No extrude or sweep, but the PDF spline has already been written to the hierarchy
            return None,0,0,0
        else:
            # Return the basic spline
            return spl,depth,offset,spread

def CreateExtrude(spl,node,depth,col,userdataIdx,chksum,oldchksum):
    ext = c4d.BaseObject(c4d.Oextrude)
    ext.SetAbsPos(spl.GetAbsPos())
    ext[c4d.ID_BASEOBJECT_USECOLOR] = c4d.ID_BASEOBJECT_USECOLOR_AUTOMATIC
    ext[c4d.ID_BASEOBJECT_COLOR] = col
    # Set Extrude Values based on ArtSmartObject
    for i,val in node.GetData():
        if i>=2000 and i<3999:
            ext[i]=val

    if c4d.GetC4DVersion() > 22600:
        ext[c4d.EXTRUDEOBJECT_DIRECTION] = c4d.EXTRUDEOBJECT_DIRECTION_Z
        ext[c4d.EXTRUDEOBJECT_EXTRUSIONOFFSET] = abs(depth)
    elif c4d.GetC4DVersion() >=21600:
        ext[c4d.EXTRUDEOBJECT_EXTRUSIONOFFSET] = -abs(depth)
    else:
        ext[c4d.EXTRUDEOBJECT_MOVE] = c4d.Vector(0,0,abs(depth))
    ext.SetName(spl.GetName())
    
    phong = ext.MakeTag(c4d.Tphong)
    phong[c4d.PHONGTAG_PHONG_ANGLELIMIT] = True
    phong[c4d.PHONGTAG_PHONG_ANGLE] = math.pi/3
    phong[c4d.PHONGTAG_PHONG_USEEDGES] = True
    
    # Hack to support the old checksum value (where multiple splines could get the same checksum)
    if oldchksum in userdataIdx and userdataIdx[oldchksum] is not None:
        chksum = oldchksum
        
    if chksum in userdataIdx and userdataIdx[chksum] is not None:
        if "Rnd" in userdataIdx[chksum] and node[userdataIdx[chksum]["Rnd"]] is not None:
            tex = c4d.TextureTag()
            tex[c4d.TEXTURETAG_MATERIAL] = node[userdataIdx[chksum]["Rnd"]]
            tex[c4d.TEXTURETAG_RESTRICTION] = "R1"
            tex[c4d.TEXTURETAG_PROJECTION] = c4d.TEXTURETAG_PROJECTION_UVW
            ext.InsertTag(tex)
            tex = c4d.TextureTag()
            tex[c4d.TEXTURETAG_MATERIAL] = node[userdataIdx[chksum]["Rnd"]]
            tex[c4d.TEXTURETAG_RESTRICTION] = "R2"
            tex[c4d.TEXTURETAG_PROJECTION] = c4d.TEXTURETAG_PROJECTION_UVW
            ext.InsertTag(tex)
        if "Cap" in userdataIdx[chksum] and node[userdataIdx[chksum]["Cap"]] is not None:
            tex = c4d.TextureTag()
            tex[c4d.TEXTURETAG_MATERIAL] = node[userdataIdx[chksum]["Cap"]]
            tex[c4d.TEXTURETAG_RESTRICTION] = "C1"
            tex[c4d.TEXTURETAG_PROJECTION] = c4d.TEXTURETAG_PROJECTION_FLAT
            tex[c4d.TEXTURETAG_SIZE] = spl.GetRad()+c4d.Vector(0,0,1)
            ext.InsertTag(tex)
            tex = c4d.TextureTag()
            tex[c4d.TEXTURETAG_MATERIAL] = node[userdataIdx[chksum]["Cap"]]
            tex[c4d.TEXTURETAG_RESTRICTION] = "C2"
            tex[c4d.TEXTURETAG_PROJECTION] = c4d.TEXTURETAG_PROJECTION_FLAT
            tex[c4d.TEXTURETAG_SIZE] = spl.GetRad()+c4d.Vector(0,0,1)
            ext.InsertTag(tex)
        if "Ext" in userdataIdx[chksum] and node[userdataIdx[chksum]["Ext"]] is not None:
            tex = c4d.TextureTag()
            tex[c4d.TEXTURETAG_MATERIAL] = node[userdataIdx[chksum]["Ext"]]
            tex[c4d.TEXTURETAG_PROJECTION] = c4d.TEXTURETAG_PROJECTION_UVW
            ext.InsertTag(tex)
            
    return ext

def CreateSweep(spl,node,col,userdataIdx,chksum,depth=0,width=0):
    #width = node[ART_STROKEWIDTH]
    if width <= 0: return None

    aspect = node[ART_STROKEASPECT]
    rounding = node[ART_STROKEROUNDING]
    height = aspect + abs(depth)
    
    #if height <= 0: return None
    
    sweep = c4d.BaseObject(c4d.Osweep)
    sweep.SetAbsPos(spl.GetAbsPos())
    sweep[c4d.ID_BASEOBJECT_USECOLOR] = c4d.ID_BASEOBJECT_USECOLOR_AUTOMATIC
    sweep[c4d.ID_BASEOBJECT_COLOR] = col
    sweep.SetName(spl.GetName())

    if node[ART_STROKEGROWTH] >= 0:
        sweep[c4d.SWEEPOBJECT_GROWTH] = node[ART_STROKEGROWTH]
        sweep[c4d.SWEEPOBJECT_STARTGROWTH] = 0
    else:
        sweep[c4d.SWEEPOBJECT_GROWTH] = 1
        sweep[c4d.SWEEPOBJECT_STARTGROWTH] = 1 + node[ART_STROKEGROWTH]
        
    
    phong = sweep.MakeTag(c4d.Tphong)
    phong[c4d.PHONGTAG_PHONG_ANGLELIMIT] = True
    phong[c4d.PHONGTAG_PHONG_ANGLE] = math.pi/3
    phong[c4d.PHONGTAG_PHONG_USEEDGES] = True
    
    profile = c4d.BaseObject(c4d.Osplinerectangle)
    profile[c4d.PRIM_RECTANGLE_WIDTH] = width
    profile[c4d.PRIM_RECTANGLE_HEIGHT] = height
    if rounding > 0:
        profile[c4d.PRIM_RECTANGLE_ROUNDING] = True
        profile[c4d.PRIM_RECTANGLE_RADIUS] = min(width,height)/2 * rounding
        if rounding == 1:
            profile[c4d.SPLINEOBJECT_INTERPOLATION] = c4d.SPLINEOBJECT_INTERPOLATION_UNIFORM
            profile[c4d.SPLINEOBJECT_SUB] = 2
        else:
            profile[c4d.SPLINEOBJECT_INTERPOLATION] = c4d.SPLINEOBJECT_INTERPOLATION_ADAPTIVE
            profile[c4d.SPLINEOBJECT_ANGLE] = math.pi/18
    spl.InsertUnder(sweep)
    profile.InsertUnder(sweep)
    
    #Set Alignment
    off = node[ART_STROKEOFFSET]
    #zalign = (depth/2)+(depth*aspect/2*off)
    #zalign = (depth*off)-(depth*aspect/2)
    strokedepth = height
    strokecenter = strokedepth/2
    center = abs(depth)/2
    zalign = center + (center*off) - (strokecenter*off)
    if node[ART_STROKEALIGN_STARTEND]:
        sp1 = sweep
        sp2 = sweep.GetClone()
        sweep = c4d.BaseObject(c4d.Onull)
        sweep.SetAbsPos(sp1.GetAbsPos())
        sp1.InsertUnder(sweep)
        sp2.InsertUnder(sweep)
        sp1.SetAbsPos(c4d.Vector(0,0,0))
        sp2.SetAbsPos(c4d.Vector(0,0,0))
        sp1.GetDownLast().SetAbsPos(c4d.Vector(0,0,(strokecenter*off)))
        sp2.GetDownLast().SetAbsPos(c4d.Vector(0,0,abs(depth)-(strokecenter*off)))
    else:
        spl.SetAbsPos(c4d.Vector(0,0,zalign))
    
    if chksum in userdataIdx and userdataIdx[chksum] is not None:    
        if "Sweep" in userdataIdx[chksum] and node[userdataIdx[chksum]["Sweep"]] is not None:
            tex = c4d.TextureTag()
            tex[c4d.TEXTURETAG_MATERIAL] = node[userdataIdx[chksum]["Sweep"]]
            tex[c4d.TEXTURETAG_PROJECTION] = c4d.TEXTURETAG_PROJECTION_UVW
            sweep.InsertTag(tex)

    return sweep
    
def _GenerateChecksum(points):
    return str(reduce(lambda x,y : x^y, [hash(repr(pt)) for pt in points]))

def GenerateChecksum(points):
    return hashlib.md5(repr(points).encode('utf-8')).hexdigest()
    
def GetUserDataValue(node,userdataIdx,key,layerName="",checksum=0):
    out = 0
    
    if not node or not userdataIdx or not key: return 0
    
    try:
        if key == "Dep":
            out += node[ART_DEPTH]
        elif key == "Pat":
            out += node[ART_PATHOFFSET]
            
        if layerName in userdataIdx and key in userdataIdx[layerName] and node[userdataIdx[layerName][key]] is not None:
            out += node[userdataIdx[layerName][key]]
        
        if checksum in userdataIdx and key in userdataIdx[checksum] and node[userdataIdx[checksum][key]] is not None:
            out += node[userdataIdx[checksum][key]]
    except:
        print("Race Condition hit. Try recreating object.")
    
    return out
    
def ParseLayer(layer,parent=None,parentLayer=None,nullType=None,node=None,userdataIdx=[],zOff=0,printOnly=False,axis="Z"):
    if not layer or not isinstance(layer, geom.Layer) or not layer.contents: return 0
    
    if printOnly and (layer.printing is False or layer.visible is False): return 0
    
    lyr = None
    if parentLayer:
        doc = parentLayer.GetDocument()
        lyr = c4d.documents.LayerObject()
        lyr.SetName(layer.name)
        lyrData = {}
        lyrData["view"] = layer.visible
        lyrData["render"] = layer.printing
        lyrData["locked"] = not layer.enabled
        if layer.color and isinstance(layer.color, geom.Paint):
            lyrData["color"] = c4d.Vector(layer.color.color[0],layer.color.color[1],layer.color.color[2])
        lyr.SetLayerData(doc,lyrData)
        lyr.InsertUnder(parentLayer)
    if parent and nullType:
        doc = parent.GetDocument()
        null = c4d.BaseObject(nullType)
        null.SetName(layer.name)
        null.InsertUnder(parent)
        if nullType == FRACTURE_ID:
            if node and layer.name in userdataIdx and "Eff" in userdataIdx[layer.name]:
                null[c4d.ID_MG_MOTIONGENERATOR_EFFECTORLIST] = node[userdataIdx[layer.name]["Eff"]]
        if lyr: null.SetLayerObject(lyr)
        parent = null
   
    centers = []
    for item in layer.contents:
        if isinstance(item, geom.Layer):
            zOff = ParseLayer(item,parent,lyr,nullType,node,userdataIdx,zOff,printOnly)
            
        elif isinstance(item, geom.Path):
                                
            path,depth,offset,spread = CreatePath(item,node,userdataIdx,layer.name,axis=axis)
            
            if path:
                if parent:
                    path.InsertUnder(parent)
                    parent.ChangeNBit(c4d.NBIT_OM1_FOLD,c4d.NBITCONTROL_SET)
                if lyr:
                    path.SetLayerObject(lyr)
                
                # have to treat sweeps different
                pos = path.GetAbsPos()
                centers.append(pos)
                if axis == "Y":
                    pos.y = zOff + offset
                elif axis == "X":
                    pos.x == zOff + offset
                else:
                    pos.z = zOff + offset
                if depth < 0:
                    pos.z += depth
                path.SetAbsPos(pos)
                zOff -= spread
                
    # Recenter Layer Null
    if parent and centers:
        if node and node[ART_LEGACYCENTER]:
            center = sum(centers)/len(centers)
        else:
            center = FindCenter(centers)
        parent.SetAbsPos(center)
        for child in parent.GetChildren():
            child.SetAbsPos(child.GetAbsPos()-center)
    
    return zOff

def FindAIClipFormat():
    for formats in range(49152,65535):
        s = ctypes.create_unicode_buffer(32)
        n = ctypes.windll.user32.GetClipboardFormatNameW(formats,s,32)
        if s.value == "Encapsulated PostScript":
            return formats
    return 0

def CamVec(cam,vec,xy):

    proj = cam.GetProjection()
    if proj == c4d.Pleft:
        plane = "zy", -1
    elif proj == c4d.Pright:
        plane = "zy", 1
    elif proj == c4d.Pfront:
        plane = "xy", 1
    elif proj == c4d.Pback:
        plane = "xy", -1
    elif proj == c4d.Ptop:
        plane = "xz", 1
    elif proj == c4d.Pbottom:
        plane = "xz", -1
    else:
        ###TODO: Plane detection from global matrix
        plane = "xy",1

    if plane[0] == "xy":
        if xy == 0:
            return vec.x * plane[1]
        else:
            return vec.y
    elif plane[0] == "zy":
        if xy == 0:
            return vec.z * plane[1]
        else:
            return vec.y
    elif plane[0] == "xz":
        if xy == 0:
            return vec.x
        else:
            return vec.z * plane[1]


def IterateHierarchy(obj, deep, fn):
    output = ''
    while obj:
        output += fn(obj)
        if obj.GetDown() and (output == '' or deep is True):
            output += IterateHierarchy(obj.GetDown(), deep, fn)
        obj = obj.GetNext()
    return output

def EdgeToSpline(obj):
    c4d.utils.SendModelingCommand(c4d.MCOMMAND_EDGE_TO_SPLINE, [obj])
    output = WriteSpline(obj.GetDown())
    obj.GetDown().Remove()
    return output

def WriteSpline(obj):
    
    eps = "" # initialize output
    f = 0.75 # tangent length multiplier for C4D beziers
    
    if not isinstance(obj, c4d.BaseObject): return eps
    # Special Handling for Cloner, Fracture, MoText Objects to Iterate their cache
    if obj.GetType() == 1018544 or \
       obj.GetType() == 1018791 or \
       obj.GetType() == 1019268:
        eps += IterateHierarchy(obj.GetCache(), True, WriteSpline)

    elif isinstance(obj, c4d.PolygonObject) and c4d.IsCommandChecked(16351) and obj.GetEdgeS():
        return EdgeToSpline(obj)

    spline = obj.GetRealSpline()        # Get the Real Spline for Primitive Spline Objects
    if not spline: return eps           # Don't think I need a type test, returns none if it can't return a spline
    
    # initialize segments
    segments = []
    for seg in range(0,spline.GetSegmentCount()):
        segments.append(spline.GetSegment(seg))
    # catch for 0 segment case
    if spline.GetSegmentCount() == 0:
        if spline.GetPointCount() == 0: return eps
        segments.append({"cnt":spline.GetPointCount(),"closed":spline.IsClosed()})
    if DEBUG:
        print(segments)

    mg = obj.GetMg()
    pts = spline.GetAllPoints()
    cur = 0

    bd = obj.GetDocument().GetActiveBaseDraw()
    cam = bd.GetSceneCamera(obj.GetDocument())
    if cam is None: cam = bd.GetEditorCamera(obj.GetDocument())

    # Get ArtSmart Tag for Stroke & Fill info (if it exists)
    tag=spline.GetTag(TAG_ID)
    close = "S"
    if tag:
        stroke = tag[ART_STROKECOLOR]
        fill = tag[ART_FILLCOLOR]
        width = tag[ART_STROKEWIDTH]
        eps += str(stroke.x) + " " + str(stroke.y) + " " + str(stroke.z) + " XA\n"
        eps += str(fill.x) + " " + str(fill.y) + " " + str(fill.z) + " Xa\n"
        if width > 0:
            eps += str(width) + " w\n"
        if tag[ART_STROKED] and tag[ART_FILLED]:
            close = "B"
        elif tag[ART_FILLED]:
            close = "F"
        elif tag[ART_STROKED]:
            close = "S"
        else:
            close = "N"
   
    if spline.IsClosed():
            close = close.lower()       # Closed splines have the lower case equivalent end char in EPS
            
    # Loop through segments, we write each separate in EPS
    for seg in range(0,len(segments)):
        segcnt = segments[seg]["cnt"]        #points in segment
        segstart = cur          #first point index of segment
        segend = cur + segcnt  #last point index of segment

        if DEBUG:
            print("="*10,"SEGMENT", seg)
            print("Cnt=",segcnt)
            print("Start=",segstart)
            print("End=",segend)
        
        # closed segments put the m at the last point, open put the m at the first point
        segclosed = segments[seg]["closed"]
        #if segclosed:
        #    midx = segend-1
        #else:
        midx = segstart
        
        # offset all points with the global matrix of the spline to put them in the right spot if we Paste in Place in AI
        eps += fstr(CamVec(cam,pts[midx]*mg,0)) + " " + fstr(CamVec(cam,pts[midx]*mg,1)) + " m\n"   #initial move
        for idx in range(segstart,segend):
            if segclosed or (idx >= segstart and idx+1 < segend):
                pidx = idx
                idx = idx + 1
                #if pidx < segstart: pidx = segend - 1
                if idx >= segend: idx = segstart
                if idx < spline.GetTangentCount():
                    tng = spline.GetTangent(idx)
                else:
                    tng = {'vl':c4d.Vector(0,0,0),'vr':c4d.Vector(0,0,0)}
                if pidx < spline.GetTangentCount():
                    ptng = spline.GetTangent(pidx)
                else:
                    ptng = {'vl':c4d.Vector(0,0,0),'vr':c4d.Vector(0,0,0)}
                ppt = pts[pidx] * mg
                pt = pts[idx] * mg
                eps += fstr(CamVec(cam,ppt,0)+(CamVec(cam,mg.MulV(ptng["vr"]),0)/f)) + " " + fstr(CamVec(cam,ppt,1)+(CamVec(cam,mg.MulV(ptng["vr"]),1)/f)) + " "
                eps += fstr(CamVec(cam,pt,0)+(CamVec(cam,mg.MulV(tng["vl"]),0)/f)) + " " + fstr(CamVec(cam,pt,1)+(CamVec(cam,mg.MulV(tng["vl"]),1)/f)) + " "
                eps += fstr(CamVec(cam,pt,0)) + " " + fstr(CamVec(cam,pt,1)) + " c\n"
            if cur != segend:
                cur += 1
            
        eps = eps + close + "\n"
    
    # Make multi-segment splines into compound paths
    if spline.GetSegmentCount() > 1:
        eps = "*u\n"+eps+"*U\n"
        
    return eps

def fstr(n):
    return str("%.3f" % n)
    
def AICBHeader():
    str = """
%!PS-Adobe-3.0 EPSF-3.0
%%Creator: Adobe Illustrator(R) 8.0
%%AI8_CreatorVersion: 16.0.2
%%Canvassize: 16383
%%BoundingBox: 119 -409 298 -140
%%HiResBoundingBox: 119.5 -408.5 297.5 -140.5
%%DocumentProcessColors: Black
%%DocumentSuppliedResources: procset Adobe_level2_AI5 1.2 0
%%+ procset Adobe_ColorImage_AI6 1.3 0
%%+ procset Adobe_Illustrator_AI5 1.3 0
%%+ procset Adobe_cshow 2.0 8
%%+ procset Adobe_shading_AI8 1.0 0
%AI5_FileFormat 4.0
%AI3_ColorUsage: Black&White
%AI3_IncludePlacedImages
%AI7_ImageSettings: 1
%%AI6_ColorSeparationSet: 1 1 (AI6 Default Color Separation Set) 
%%+ Options: 1 16 0 1 1 1 0 0 0 0 1 1 1 18 0 0 0 0 0 0 0 0 -1 -1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 0 1 2 3 4
%%+ PPD: 1 21 0 0 60 45 2 2 1 0 0 1 0 0 0 0 0 0 0 0 1 0 () 
%AI3_TemplateBox: 306.5 -396.5 306.5 -396.5
%AI3_TileBox: 0 -792 612 0
%AI3_DocumentPreview: Header
%AI5_ArtSize: 612 792
%AI5_RulerUnits: 2
%AI5_ArtFlags: 0 0 0 1 0 0 1 0 0
%AI5_TargetResolution: 800
%AI5_NumLayers: 1
%%PageOrigin:0 -792
%AI7_GridSettings: 72 8 72 8 1 0 0.8 0.8 0.8 0.9 0.9 0.9
%AI9_Flatten: 0
%AI12_CMSettings: 00.MS
%%EndComments
%%BeginProlog
%%BeginResource: procset Adobe_level2_AI5 1.2 0
%%Title: (Adobe Illustrator (R) Version 5.0 Level 2 Emulation)
%%Version: 1.2 0
%%CreationDate: (04/10/93) ()
%%Copyright: ((C) 1987-1996 Adobe Systems Incorporated All Rights Reserved)
%%EndProlog
%%BeginSetup
%AI5_Begin_NonPrinting
Np
%AI5_End_NonPrinting--
%AI5_BeginPalette
0 0 Pb
PB
%AI5_EndPalette
%%EndSetup
%AI5_BeginLayer
1 1 1 1 0 0 1 0 79 128 255 0 50 Lb
(Layer 1) Ln
0 A
1 Ap
%AI3_Note:<< /operation /union >>
0 O
0 0 0 0 k
0 R
0 0 0 1 K
0 J 0 j 1 w 10 M []0 d
0 XR
    """            
    return str
        
def AICBFooter():
    str = """
LB
%AI5_EndLayer--
%%PageTrailer
gsave annotatepage grestore showpage
%%Trailer
Adobe_Illustrator_AI5 /terminate get exec
Adobe_shading_AI8 /terminate get exec
Adobe_ColorImage_AI6 /terminate get exec
Adobe_cshow /terminate get exec
Adobe_level2_AI5 /terminate get exec
%%EOF
    """
    return str
    